-- pod.ex - Plain Old Documentation (POD) generator
-- version 1.02
-- 8.20.99 - Buddy Hyllberg <budmeister1@juno.com>

---------------
-- Changes:
--
-- v1.02
-- * added support for loadable generator modules (and fixed do_options())
-- * renamed build_outfile() to strip_ext()
-- * made pod.ex work more on the PODs before passing to the generators
-- * rewrote extract_pod() with a better (and more flexible) search algorithm
-- * added check_mod_return()
-- * added -m (list modules) option and list_modules()
--
-- v1.00
-- * first wrote pod!
-- * only support plain text; v1.02 will generate HTML
-- * made it stop gagging on =commands with no text
-- * fixed a few bugs
-- * toyed with idea of using dynamic include files... extendable output format support?
--
---------------

include pods.e        -- list of loadable modules
include wildcard.e

--with trace -- trace(1)
without warning
without type_check

constant VERSION = "Plain Old Documentation (POD) generator\nversion 1.02 - Buddy Hyllberg <budmeister1@juno.com>\n",
         USAGE = "\nUsage: pod [-?] [-m] [-f<format>] [-o<outfile>] program.ex\n\nNotes:\n" &
                 "- Extensions are *not* assumed.\n" &
                 "- If no output file is given with -o, a file with the same name is created.\n" &
                 "- Output format defaults to plain text if none is given using -f.\n" &
                 "- To list the installed output formats, use -m.\n" &
                 "- If the format isn't included in pods.e, it won't be recognized as valid.\n"


sequence OUTFILE, INFILE, GEN_SELECTED
INFILE = {}   OUTFILE = {}   GEN_SELECTED = {}


procedure die( sequence msg )
  puts(1, msg)
  abort(1)
end procedure

function failed( object x )
  if equal(x, -1) or equal(x, 0) then
    return 1
  else
    return 0
  end if
end function


function read( sequence filename )
  atom IN
  object buf, lin
  buf = {}   lin = {}
  IN = open(filename, "r")
  if failed(IN) then  return -1  end if
  while 1 do
    lin = gets(IN)
    if atom(lin) then exit end if
    if lin[length(lin)] = '\n' then
      lin = lin[1..length(lin)-1]  -- strip \n
    end if
    if length(lin) then
      buf = append(buf, lin)  -- don't add blank lines
    end if
  end while
  close(IN)
  return buf
end function


-- this is given a full path to a file. it's job is to strip
-- the file extension (if there is one).
function strip_ext( sequence origname )
  sequence x   x = {}
  for i = length(origname) to 1 by -1 do  -- look for the extension starting at the end
    if equal(origname[i], '.') then
      if match("\\", origname[i..length(origname)]) then -- check to see if it's past the filename
        return origname                                  -- ...yup.
      else
        x = origname[1..i-1]
        return x
      end if
    end if
  end for
end function


-- lists the currently installed output modules
procedure list_modules()
  puts(1, "\nHere are the current generators:\n")
  for q = 1 to length(GENERATORS) do
    puts(1, GENERATORS[q] & "\t")
  end for
end procedure


-- this sets switches according to options given.
procedure do_options( sequence argv ) -- gosh I luv C ;)
  -- Wow! total rewrite in v1.02 *whew*
  if length(argv) = 1 then                   -- only 1 argument
    if wildcard_file("*.e*", argv[1]) then -- ummm... it's a file, right?
      INFILE = argv[1]
    elsif equal(argv[1], "-?") then        -- they need help ;)
      die(VERSION & USAGE)
    elsif equal(argv[1], "-m") then        -- they want the installed modules
      list_modules()
      die("\n")
    else
      die(USAGE)
    end if
  else                                       -- 2+ arguments
    for i = 1 to length(argv) do
      -- remember the usage boys & girls!
      -- pod [-?] [-f<format>] [-o<outfile>] program.ex
      if equal(argv[i], "-?") then  -- give em help, even if they had other args
        die(VERSION & USAGE)
      elsif wildcard_file("*.e*", argv[i]) then  -- haha, we got a filename
        if length(INFILE) then  -- did we already get a filename?
          die("Oops: multiple input files are not allowed.\n" & USAGE)
        else
          INFILE = argv[i]
        end if
      elsif equal(argv[i], "-m") then  -- they want a list of modules
        list_modules()
        die("\n")
      elsif wildcard_match("-f?*", argv[i]) then -- use the ? to make sure there's a filename included
        if find(argv[i][3..length(argv[i])], GENERATORS) then  -- it's plugged in
          GEN_SELECTED = argv[i][3..length(argv[i])]
        else                                                   -- ...or not...
          puts(1, "Oops: output generator isn't recognized.\n")
          list_modules()
          die("\n")
        end if
      elsif wildcard_match("-o?*", argv[i]) then -- dittos
        if length(OUTFILE) then
          die("Oops: multiple output files are not allowed.\n" & USAGE)
        else
          OUTFILE = argv[i][3..length(argv[i])]
        end if
      else
        die(USAGE)
      end if
    end for
  end if
end procedure


-- this separates the POD docs from the real comments. it doesn't do
-- any translation or parsing of PODs, just finding. save that for
-- the generators :)
function extract_pod( sequence code )
  atom where_comment, where_command
  object currentln, podbuf
  podbuf = {}
  for i = 1 to length(code) do
    currentln = code[i]
    where_comment = match("--", currentln) -- try to find a comment
    if where_comment then  -- found a match!
      -- grab the =command to the end of the line (if found)
      where_command = match("=", currentln[where_comment+2..length(currentln)])
      if where_command then
        -- extract the '=command text' part :)
        podbuf = append(podbuf, currentln[where_comment+2+where_command-1..length(currentln)])
      end if
    end if
  end for
  if length(podbuf) then -- make sure it found something
    return podbuf
  else
    return -1
  end if
end function


-- this separates the =commands from their text. it returns a sequence like this:
-- {{"=command", "text"},{"=command", "text"}, ...}
-- that makes it easier for the generators to work.
function split_commands( sequence pods )

  sequence buf, line, command, text
  buf = {}

  for i = 1 to length(pods) do

    line = pods[i]

    if failed(match(" ", line)) then  -- is there only a command and no text?
      buf = append(buf, {line,{}})

    elsif wildcard_match("=* ", line) then  -- is there only a command + a space? (still no text)
      buf = append(buf, {line[1..length(line)-1],{}}) -- yeah, just do the command again

    else  -- looks like it's normal... [famous last words]
      command = line[1..match(" ", line)-1]          -- grab the command
      text = line[match(" ", line)+1..length(line)]  -- and the text*
      buf = append(buf, {command, text})             -- add it to the list

      -- *Note: assumes the text starts _immediately_ after the =command (one space after).
      -- Any extra spaces are taken literally (ie: included in the text).

    end if

  end for

  return buf

end function


procedure write( sequence filename, sequence data )
  atom OUT
  OUT = open(filename, "w")
  if failed(OUT) then  die(sprintf("Oops: can't write to %s\n", {filename}))  end if
  if length(data) = 1 then
    puts(OUT, data)
  else
    for i = 1 to length(data) do
      puts(OUT, data[i])
    end for
  end if
  close(OUT)
end procedure


-- I exist only to check for certain things: length, extension, and length.
-- What a life :)
procedure check_mod_return( sequence checkme, sequence module_name )
  if length(checkme) != 2 then  -- make sure it's 2 elements (remember: {ext, data})
    die(sprintf("Oops in module [%s]: Invalid return data (missing extension or data)\n", {module_name}))
  end if
  if failed(wildcard_file(".*", checkme[1])) then  -- it's not a valid extension
    die(sprintf("Oops in module [%s]: Invalid return data (bad filename extension)\n", {module_name}))
  end if
  if failed(length(checkme[2])) then  -- data is empty... that's a no-no :)
    die(sprintf("Oops in module [%s]: Invalid return data (no data to write)\n", {module_name}))
  end if
end procedure


procedure main() ---------------------- main procedure -----------------------

  object pod, ext
  sequence x, z

  x = lower(command_line())  x = x[3..length(x)] -- get just the arguments
  if length(x) then
    do_options(x)
  else
    die(USAGE)
  end if

  pod = read(INFILE)      -- read the source code
  if failed(pod) then  die(sprintf("Oops: can't read from %s\n", {INFILE}))  end if

  pod = extract_pod(pod)  -- extract the pod source
  if failed(pod) then  die(sprintf("Oops: no PODs found in %s\n", {INFILE}))  end if

  pod = split_commands(pod)  -- separate the =commands from their text

  -- next figure out what our output file is going to be
  if length(OUTFILE) then
    OUTFILE = strip_ext(OUTFILE)
  else
    OUTFILE = strip_ext(INFILE)
  end if

  -- now decide what generator to use & write the data
  if length(GEN_SELECTED) then
    z = call_func(routine_id("gen_" & GEN_SELECTED), {pod})  -- run the specified generator
    check_mod_return(z, GEN_SELECTED)  -- if it comes back here, then everything was ok
    write(OUTFILE & z[1], z[2])  -- everything looks ok, so go ahead and write it

  else
    z = call_func(routine_id("gen_text"), {pod}) -- defaults to plain text
    check_mod_return(z, GEN_SELECTED)
    write(OUTFILE & z[1], z[2])

  end if

end procedure



main() -- pod start
